# profiler.tcl --
#
#	Tcl code profiler.
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @ (#) $ Id: profiler.tcl,v 1.12 2000/09/20 17:23:14 ericm Exp $

# ==============================================================================
# sgxCP.tcl --
#
#	sgxCP: SoftGuards eXtension for Code Profiling
#
# Modified to support tclsh8.0 and later (string equal ..., 
#	clock clicks/1000 - milliseconds)
# Modified ::sgxCP::print output format of descendants 
#	and list never used proc(s)
# Code was formatted using frink and modified to pass its heuristics.
#
# 	changes copyright (c) 2000 by Stefan G. R. Finzel <finzel@osn.de>.
#
# RCS: @(#) $Id: sgxCP.tcl,v 1.7 2002/11/08 18:31:08 fis Exp $
# ==============================================================================

package provide sgxCP 1.0

namespace eval ::sgxCP:: {
    namespace export init

    variable enabled 1
}

# ::sgxCP::tZero --
#
#	Start a named timer instance
#
# Arguments:
#	tag	name for the timer instance; if none is given, defaults to ""
#
# Results:
#	None.

proc ::sgxCP::tZero {{ tag "" }} {
    set ms [expr {[clock clicks] / 1000}]
    set us [clock clicks]
    regsub -all -- {:} ${tag} {} tag
    # FRINK: nocheck
    set "::sgxCP::T${tag}" [list ${us} ${ms}]
    return
}

# ::sgxCP::tMark --
#
#	Return the delta time since the start of a named timer.
#
# Arguments:
#	tag	Tag for which to return a delta; if none is given, defaults to
#		"" 
#
# Results:
#	dt	Time difference between start of the timer and the current
#		time, in clicks/1000 (not microseconds).

proc ::sgxCP::tMark {{ tag "" }} {
    set ut [clock clicks]
    set mt [expr {[clock clicks] / 1000}]
    regsub -all -- {:} ${tag} {} tag
    # FRINK: nocheck
    set ust [lindex [set ::sgxCP::T${tag}] 0]
    # FRINK: nocheck
    set mst [lindex [set ::sgxCP::T${tag}] 1]
    set udt [expr {(${ut}-${ust})}]
    set mdt [expr {(${mt}-${mst})}]000
    set dt ${udt}
    ## handle wrapping of the click/1000 (not microsecond) clock
    if {${dt} < 0 || ${dt} > 1000000} {
        set dt ${mdt}
    }
    set dt
}

# ::sgxCP::stats --
#
#	Compute statistical information for a set of values, including
#	the mean, the standard deviation, and the covariance.
#
# Arguments:
#	args	Values for which to compute information.
#
# Results:
#	A list with three elements:  the mean, the standard deviation, and the
#	covariance.

proc ::sgxCP::stats {args} {
    set sum 0
    set mean 0
    set sigma 0
    set cov 0
    set N [llength ${args}]
    if {${N} > 0} {
        foreach val ${args} {
            set sum [expr {${sum}+${val}}]
        }
        set mean [expr {${sum}/${N}}]
    }
    if {${N} > 1} {
        set sigma_sq 0
        foreach val ${args} {
            set sigma_sq [expr {${sigma_sq} + pow((${val} - ${mean}) , 2)}]
        }
        set sigma_sq [expr {${sigma_sq} /(${N} - 1)}]
        set sigma [expr {round(sqrt(${sigma_sq}))}]
        set cov [expr {((${sigma} * 1.0) / ${mean}) * 100}]
        set cov [expr {round(${cov} * 10) / 10.0}]
    }
    return [list ${mean} ${sigma} ${cov}]
}

# ::sgxCP::Handler --
#
#	Profile a function.  This function works together with profProc, which
#	replaces the proc command.  When a new procedure is defined, it creates
#	and alias to this function; when that procedure is called, it calls
#	this handler first, which gathers profiling information from the call.
#
# Arguments:
#	name	name of the function to profile.
#	args	arguments to pass to the original function.
#
# Results:
#	res	result from the original function.

proc ::sgxCP::Handler {name args} {
    variable enabled
    if {${enabled}} {
        variable callers
        if {[info level] == 1} {
            set caller [info script]
        } else {
            # Get the name of the calling procedure
            set caller [lindex [info level -1] 0]
            # Remove the ORIG suffix
            set caller [string range ${caller} 0 [expr {[string length ${caller}] - 5}]]
        }
        if {[catch {incr callers(${name},${caller})}]} {
            set callers(${name},${caller}) 1
        }
        ::sgxCP::tZero ${name}.${caller}
    }

    set CODE [uplevel ${name}ORIG ${args}]
    if {${enabled}} {
        set t [::sgxCP::tMark ${name}.${caller}]
        variable statTime
        lappend statTime(${name}) ${t}

        variable callCount
        if {[incr callCount(${name})] == 1} {
            variable compileTime
            set compileTime(${name}) ${t}
        }
        variable totalRuntime
        incr totalRuntime(${name}) ${t}
        variable descendantTime
        if {[catch {incr descendantTime(${caller}) ${t}}]} {
            set descendantTime(${caller}) ${t}
        }
        variable descendants
        if {[catch {incr descendants(${caller},${name})}]} {
            set descendants(${caller},${name}) 1
        }
    }
    return ${CODE}
}

# ::sgxCP::profProc --
#
#	Replacement for the proc command that adds rudimentary profiling
#	capabilities to Tcl.
#
# Arguments:
#	name		name of the procedure
#	arglist		list of arguments
#	body		body of the procedure
#
# Results:
#	None.

proc ::sgxCP::profProc {name arglist body} {
    variable callCount
    variable compileTime
    variable totalRuntime
    variable descendantTime
    variable statTime

    # Get the fully qualified name of the proc
    set ns [uplevel [list namespace current]]
    # If the proc call did not happen at the _global_ context and it did not
    # have an absolute namespace qualifier, we have to prepend the current
    # namespace to the command name
    if {[string compare ${ns} "::"]} {
        if {![regexp {^::} ${name} UNUSED]} {
            set name "${ns}::${name}"
        }
    }
    if {![regexp {^::} ${name} UNUSED]} {
        set name "::${name}"
    }

    # Set up accounting for this procedure
    set callCount(${name}) 0
    set compileTime(${name}) 0
    set totalRuntime(${name}) 0
    set descendantTime(${name}) 0
    set statTime(${name}) {}

    uplevel 1 [list ::_oldProc ${name}ORIG ${arglist} ${body}]
    uplevel 1 [list interp alias {} ${name} {} ::sgxCP::Handler ${name}]
    return
}

# ::sgxCP::init --
#
#	Initialize the sgxCP.
#
# Arguments:
#	None.
#
# Results:
#	None.  Renames proc to _oldProc and sets an alias for proc to 
#		sgxCP::profProc

proc ::sgxCP::init {} {
    rename ::proc ::_oldProc
    interp alias {} proc {} ::sgxCP::profProc

    return
}

# ::sgxCP::print --
#
#	Print information about a proc.
#
# Arguments:
#	pattern	pattern of the proc's to get info for; default is *.
#
# Results:
#	A human readable printout of info.

proc ::sgxCP::print {{pattern *}} {
    variable callCount
    variable compileTime
    variable totalRuntime
    variable descendantTime
    variable descendants
    variable statTime
    variable callers

    set result ""
    set unused_proc ""
    foreach name [lsort [array names callCount ${pattern}]] {
        if {!$callCount(${name})} {
            lappend unused_proc ${name}
            continue
        }
        set avgRuntime 0
        set sigmaRuntime 0
        set covRuntime 0
        set avgDesTime 0
        if {$callCount(${name}) > 0} {
            foreach {m s c} [eval ::sgxCP::stats $statTime(${name})] {
                break
            }
            set avgRuntime ${m}
            set sigmaRuntime ${s}
            set covRuntime ${c}
            set avgDesTime [expr {$descendantTime(${name})/$callCount(${name})}]
        }

        append result "Profiling information for ${name}\n"
        append result "-------------------------\n"
        append result "            Total calls:  $callCount(${name})\n"
        append result "    Caller distribution:\n"
        foreach index [lsort [array names callers ${name},*]] {
            append result "\t\t\t  $callers(${index})\t[lindex [split ${index} {,}] end]\n"
        }
        append result "           Compile time:  $compileTime(${name})\n"
        append result "          Total runtime:  $totalRuntime(${name})\n"
        append result "        Average runtime:  ${avgRuntime}\n"
        append result "          Runtime StDev:  ${sigmaRuntime}\n"
        append result "         Runtime cov(%):  ${covRuntime}\n"
        append result "  Total descendant time:  $descendantTime(${name})\n"
        append result "Average descendant time:  ${avgDesTime}\n"
        append result "Descendants:\n"
        if {[llength [array names descendants ${name},*]]} {
            foreach index [lsort [array names descendants ${name},*]] {
                append result "\t\t\t  $descendants(${index})\t[lindex [split ${index} {,}] end]\n"
            }
        } else {
            append result "  none\n"
        }
        append result "\n"
    }
    if {[llength ${unused_proc}]} {
        append result "Never used:\t[llength ${unused_proc}] proc(s)\n\t[join ${unused_proc}\
          \n\t]\n"
    }
    return ${result}
}

# ::sgxCP::dump --
#
#	Dump out the information for a proc in a big blob.
#
# Arguments:
#	pattern	pattern of the proc's to lookup; default is *.
#
# Results:
#	data	data about the proc's.

proc ::sgxCP::dump {{pattern *}} {
    variable callCount
    variable compileTime
    variable totalRuntime
    variable callers
    variable descendantTime
    variable descendants
    variable statTime

    set result {}
    # FRINK: set thisCallers
    foreach name [lsort [array names callCount ${pattern}]] {
        set i [expr {[string length ${name}] + 1}]
        catch {unset thisCallers}
        foreach index [lsort [array names callers ${name},*]] {
            set thisCallers([string range ${index} ${i} end]) $callers(${index})
        }
        set avgRuntime 0
        set sigmaRuntime 0
        set covRuntime 0
        set avgDesTime 0
        if {$callCount(${name}) > 0} {
            foreach {m s c} [eval ::sgxCP::stats $statTime(${name})] {
                break
            }
            set avgRuntime ${m}
            set sigmaRuntime ${s}
            set covRuntime ${c}
            set avgDesTime [expr {$descendantTime(${name})/$callCount(${name})}]
        }
        set descendantList [list]
        foreach index [lsort [array names descendants ${name},*]] {
            lappend descendantList [string range ${index} ${i} end]
        }
        lappend result ${name} [list callCount $callCount(${name}) callerDist [array get\
          thisCallers] compileTime $compileTime(${name}) totalRuntime $totalRuntime(${name})\
          averageRuntime ${avgRuntime} stddevRuntime ${sigmaRuntime} covpercentRuntime\
          ${covRuntime} descendantTime $descendantTime(${name}) averageDescendantTime\
          ${avgDesTime} descendants ${descendantList}]
        # FRINK: ${thisCallers}
    }
    return ${result}
}

# ::sgxCP::sortFunctions --
#
#	Return a list of functions sorted by a particular field and the
#	value of that field.
#
# Arguments:
#	field	field to sort by
#
# Results:
#	slist	sorted list of lists, sorted by the field in question.

proc ::sgxCP::sortFunctions {{field ""}} {
    switch -glob -- ${field} {
        "calls" {
                upvar ::sgxCP::callCount data
            }
        "compileTime" {
                upvar ::sgxCP::compileTime data
            }
        "totalRuntime" {
                upvar ::sgxCP::totalRuntime data
            }
        "avgRuntime" -
        "averageRuntime" {
                variable callCount
                variable totalRuntime
                foreach fxn [array names callCount] {
                    if {$callCount(${fxn}) > 1} {
                        set data(${fxn}) [expr {$totalRuntime(${fxn})/($callCount(${fxn}) - 1)}]
                    }
                }
            }
        "exclusiveRuntime" {
                variable totalRuntime
                variable descendantTime
                foreach fxn [array names totalRuntime] {
                    set data(${fxn}) [expr {$totalRuntime(${fxn}) - $descendantTime(${fxn})}]
                }
            }
        "avgExclusiveRuntime" {
                variable totalRuntime
                variable callCount
                variable descendantTime
                foreach fxn [array names totalRuntime] {
                    if {$callCount(${fxn})} {
                        set data(${fxn}) [expr {($totalRuntime(${fxn}) - $descendantTime(${fxn}))\
                          / $callCount(${fxn})}]
                    }
                }
            }
        "nonCompileTime" {
                variable compileTime
                variable totalRuntime
                foreach fxn [array names totalRuntime] {
                    set data(${fxn}) [expr {$totalRuntime(${fxn}) - $compileTime(${fxn})}]
                }
            }
        default {
                error "unknown statistic \"${field}\": should be calls, compileTime,\
                  exclusiveRuntime, nonCompileTime, totalRuntime, avgExclusiveRuntime, or\
                  avgRuntime"
            }
    }

    set result [list]
    foreach fxn [array names data] {
        lappend result [list ${fxn} $data(${fxn})]
    }
    return [lsort -integer -index 1 ${result}]
}

# ::sgxCP::reset --
#
#	Reset collected data for functions matching a given pattern.
#
# Arguments:
#	pattern		pattern of functions to reset; default is *.
#
# Results:
#	None.

proc ::sgxCP::reset {{pattern *}} {
    variable callCount
    variable compileTime
    variable totalRuntime
    variable callers
    variable statTime

    foreach name [array names callCount ${pattern}] {
        set callCount(${name}) 0
        set compileTime(${name}) 0
        set totalRuntime(${name}) 0
        set statTime(${name}) {}
        foreach caller [array names callers ${name},*] {
            unset callers(${caller})
        }
    }
    return
}

