# ==============================================================================
# sgxCT.tcl --
#
#	SoftGuards eXtension Code Tracer.
#
# Copyright (c) 2000 by T&E Soft, Stefan G. R. Finzel <finzel@osn.de>.
# All rights reserved.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @(#) $Id: sgxCT.tcl,v 1.14 2002/10/04 10:17:29 fis Exp $
# ==============================================================================
# This module bases on the idea of:
#
# profiler.tcl --
#
#	Tcl code profiler.
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @ (#) $ Id: profiler.tcl,v 1.12 2000/09/20 17:23:14 ericm Exp $
# ==============================================================================


package provide sgxCT 1.0

namespace eval ::sgxCT:: {
    namespace export Init

    variable sgxCT
    variable count

    # element      | description
    # ------------ | -----------
    # enabled        0/1=de-/activate code trace
    # list_called    0/1=do not/ list all called procedures matching a regular expression
    # list_caller    0/1=do not/ list all caller procedures matching a regular expression
    # list_callings  0/1=do not/ list all calling pairs of procedures matching a regular expression
    # list_calls     0/1=do not/ list all unused procedures matching a regular expression
    # ------------
    # init           internal use only: package 0/1=not/ initialized 
    # fh             internal use only: an open filehandle to print output to
    # regexp         internal use only: regular expression to create trace information
    array set sgxCT {
      enabled       1 
      list_called   1 
      list_caller   1 
      list_callings 1 
      list_calls    1
      
      init          0 
      fh            stderr
      regexp        {.*}
    }
}

# ------------------------------------------------------------------------------
# ::sgxCT::Handler --
#
#	trace a function.  This function works together with traceProc, which
#	replaces the proc command.  When a new procedure is defined, it creates
#	and alias to this function; when that procedure is called, it calls
#	this handler first, which gathers tracing information from the call.
#
# Arguments:
#	name	name of the function to trace.
#	args	arguments to pass to the original function.
#
# Results:
#	res	result from the original function.
# ------------------------------------------------------------------------------
proc ::sgxCT::Handler {name args} {
    # --------------------------------------------------------------------------
    variable sgxCT
    variable count

    if {$sgxCT(enabled)} {
        if {[info level] == 1} {
            set caller [info script]
        } else {
            # Get the name of the callings procedure
            set caller [lindex [info level -1] 0]
            # Remove the SGXCT suffix
            set caller [string range ${caller} 0 [expr {[string length ${caller}] - 6}]]
        }
        if {[regexp -- $sgxCT(re) ${name} UNUSED]} {
            if {[catch {incr count(called:${name})}]} {
                set count(called:${name}) 1
            }
            if {[catch {incr count(caller:${caller})}]} {
                set count(caller:${caller}) 1
            }
            if {[catch {incr count(callings:${caller},${name})}]} {
                set count(callings:${caller},${name}) 1
            }
            incr count(calls:${name})
            puts $sgxCT(fh) "call\t[info level]\t${name}\t${args}"
        } elseif {[regexp -- $sgxCT(re) ${caller}]} {
            if {[catch {incr count(caller:${caller})}]} {
                set count(caller:${caller}) 1
            }
            if {[catch {incr count(callings:${caller},${name})}]} {
                set count(callings:${caller},${name}) 1
            }
        }
    }
    set res [uplevel ${name}SGXCT ${args}]
    if {$sgxCT(enabled) && [regexp -- $sgxCT(re) ${name} UNUSED]} {
        puts $sgxCT(fh) "return\t[info level]\t${name}\t${res}"
    }
    return ${res}
}

# ------------------------------------------------------------------------------
# ::sgxCT::Init --
#
#	Initialise the traceing.
#
# Arguments:
#	re        regular expression selecting the procs to trace; default is all {.*}.
#       fh        an open file handle to use for output; default is stderr
#
# Results:
#	None.  Renames ::proc to ::procSGXCT and sets an alias for proc to 
#		::sgxCT::traceProc
# Side effetcs:
#       Reset all prior settings and collected data
# ------------------------------------------------------------------------------
proc ::sgxCT::Init {{re {}} {fh stderr}} {
    # --------------------------------------------------------------------------
    variable sgxCT

    if {!$sgxCT(init) && [llength ${re}]} {
        set sgxCT(init) 1
        rename ::proc ::procSGXCT
        interp alias {} proc {} ::sgxCT::traceProc
    }
    ::sgxCT::Reset
    set sgxCT(re) ${re}
    set sgxCT(fh) ${fh}
    if {[llength $sgxCT(re)]} {
        puts $sgxCT(fh) "# Call Trace\tregexp: '$sgxCT(re)'\n\nmode\tlevel\tproc\targs/return"
    }
    return
}

# ------------------------------------------------------------------------------
# ::sgxCT::Print --
#
#	Print information about a proc.
#
# Arguments:
#	re		pattern of the proc's to get info for; default is {.*}.
#
# Results:
#	A human readable printout of info.
# ------------------------------------------------------------------------------
proc ::sgxCT::Print {{re {.*}}} {
    # --------------------------------------------------------------------------
    variable sgxCT
    variable count

    set result "\n# Call Trace Summary\tregexp: '${re}'"
    foreach group {called: caller: callings: calls:} {
        if {$sgxCT(list_[string trim ${group} {:}])} {
            if {[string compare ${group} {calls:}]} {
                append result "\n\n\t# ${group}"
            } else {
                append result "\n\n\t# unused:"
            }
            foreach name [lsort -dictionary [array names count ${group}*]] {
                if {![string compare ${group} {calls:}]} {
                    if {$count(${name})} {
                        continue
                    }
                }
                regsub -- ${group} ${name} {} call
                if {[regexp -- ${re} ${call} UNUSED]} {
                    set call [split ${call} {,}]
                    set tab "\t"
                    if {[string length [lindex ${call} 0]] < 8} {
                        set tab "${tab}${tab}"
                    }
                    append result "\n\t$count(${name})\t[lindex ${call} 0]${tab}[lindex ${call} 1]"
                }
            }
        }
    }
    append result "\n\n"
    return ${result}
}

# ------------------------------------------------------------------------------
# ::sgxCT::Reset --
#
#	Reset collected data for functions matching a given pattern.
#
# Arguments:
#	re		pattern of functions to reset; default is {.*}.
#
# Results:
#	None.
# ------------------------------------------------------------------------------
proc ::sgxCT::Reset {{re {.*}}} {
    # --------------------------------------------------------------------------
    variable count

    foreach group {called: caller: callings: calls:} {
        foreach name [lsort -dictionary [array names count ${group}*]] {
            regsub -- ${group} ${name} {} call
            if {[regexp -- ${re} ${call} UNUSED]} {
                if {[string compare ${group} {calls:}]} {
                    unset count(${name})
                } else {
                    set count(${name}) 0
                }
            }
        }
    }
}

# ------------------------------------------------------------------------------
# ::sgxCT::Summary --
#
#	Print summary to stderr.
#
# Arguments:
#	None.
#
# Results:
#	None.
# ------------------------------------------------------------------------------
proc ::sgxCT::Summary {{re {.*}}} {
    # --------------------------------------------------------------------------
    variable sgxCT

    puts $sgxCT(fh) [Print ${re}]
}

# ------------------------------------------------------------------------------
# ::sgxCT::traceProc --
#
#	Replacement for the proc command that adds rudimentary tracing
#	capabilities to Tcl.
#
# Arguments:
#	name		name of the procedure
#	arglist		list of arguments
#	body		body of the procedure
#
# Results:
#	None.
# ------------------------------------------------------------------------------
proc ::sgxCT::traceProc {name arglist body} {
    # --------------------------------------------------------------------------
    variable count

    # Get the fully qualified name of the proc
    set ns [uplevel [list namespace current]]
    # If the proc call did not happen at the :: context (global) and it did not
    # have an absolute namespace qualifier, we have to prepend the current
    # namespace to the command name
    if {[string compare ${ns} "::"]} {
        if {![regexp -- {^::} ${name} UNUSED]} {
            set name "${ns}::${name}"
        }
    }
    if {![regexp -- {^::} ${name} UNUSED]} {
        set name "::${name}"
    }

    # marking unused procedures
    set count(calls:${name}) 0

    uplevel 1 [list ::procSGXCT ${name}SGXCT ${arglist} ${body}]
    uplevel 1 [list interp alias {} ${name} {} ::sgxCT::Handler ${name}]
    return
}

# ==============================================================================

# example of usage :

# ::sgxCT::Init {.*}; source [file join .. test test_ct.dat]; ::sgxCT::Summary; ::sgxCT::Reset

# ==============================================================================

