# ==============================================================================
# sgxRSM.tcl --
#
#	sgxRSM: SoftGuards eXtension for Resource Standard Metrics
#
# Copyright (c) 1999 by T&E Soft, Stefan G. R. Finzel <finzel@osn.de>.
# All rights reserved.
#
# See the file "license.terms" for information on usage and redistribution of
# this file, and for a DISCLAIMER OF ALL WARRANTIES. 
#
# RCS: @(#) $Id: sgxRSM.tcl,v 1.10 2002/11/08 18:31:08 fis Exp $
# ==============================================================================
# code has to be validated and beautified to get best results
#        use e.g.: TclPro, frink
# ==============================================================================

# sgxRSM configuration:
#   tests        tests to perform (do not alter, internal use only!)
#   summary      0/1=de-/activate summary for all procs
#   LKM,proc     0/1=de-/activate language keyword metrics (calls) per proc
#   LKM,total    0/1=de-/activate language keyword metrics (calls) summary
#   rules,mean   list of rules to apply for mean value of al procs
#   rules,proc   list of rules to apply for each proc
#   rules,total  list of rules to apply for summary of all procs

# each element of a rule list is a list itself, containing 
#   an interpretable expression and a description text.
array set sgxRSM {
    tests        {_rsm_}
    LKM,proc     1
    LKM,total    1
    rules,mean   {}
    rules,proc   {}
    rules,total  {}
    summary      1
}

# support old tcl versions too
if {8.0 <= ${::tcl_version}} {
    package provide sgxRSM 1.0

    namespace eval ::sgxRSM:: {
        namespace export Init _Files_ _Pkgs_
    }
}

# ==============================================================================

# TODO: use interp

# ------------------------------------------------------------------------------
# ::sgxRSM::Init --
#                      initialize package by sourcing a list of files
# Arguments:
#        filenames:    list of filenames to source;
# Results:
#        integer       total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxRSM::Init {{filenames {}}} {
    # --------------------------------------------------------------------------
    global sgxRSM
    # FRINK: $sgxRSM() used within sourced files

    set fail 0
    foreach filename ${filenames} {
        if {[catch {source ${filename}} err]} {
            puts stderr ${err}
            incr fail
        }
    }
    return ${fail}
}

# ------------------------------------------------------------------------------
# ::sgxRSM::_Files_ --
#                      test all procedures within a list of files
# Arguments:
#        filenames:    list of filenames to source and test;
#        cmds:         list of procs to test;
# Results:
#        integer       total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxRSM::_Files_ {{filenames {}} {cmds {}}} {
    # --------------------------------------------------------------------------
    global sgxRSM

    set cmds [info commands]
    foreach filename [lsort ${filenames}] {
        puts stdout ${filename}
        if {[catch {source ${filename}} err]} {
            puts stderr ${err}
        }
    }

    # determine new commands
    set scan_cmds {}
    foreach cmd [info commands] {
        if {[lsearch -exact ${cmds} ${cmd}] < 0} {
            lappend scan_cmds ${cmd}
        }
    }
    set cmds [lsort ${scan_cmds}]
    # check procs now
    set fail 0
    foreach cmd $sgxRSM(tests) {
        incr fail [::sgxRSM::${cmd} ${cmds} [concat ${cmds} [info commands]]]
    }
    return ${fail}
}

# ------------------------------------------------------------------------------
# ::sgxRSM::_Pkgs_ --
#                     test all matching packages
# Arguments:
#        pkgs_re:     regular expression to select all packages to test
# Results:
#        integer      total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxRSM::_Pkgs_ {{pkgs_re {.*}}} {
    # --------------------------------------------------------------------------
    global prg tcl_version sgxRSM

    update
    set pkgs {}
    # FRINK: set prg(app,name); to get and add the main programs package name every time
    catch {lappend pkgs $prg(app,name)}
    # do not: package provide a_bogus_package_name or proc ::a_bogus_package_name::
    catch {package require a_bogus_package_name}
    foreach pkg [package names] {
        if {[regexp -- ${pkgs_re} ${pkg} UNUSED]} {
            lappend pkgs ${pkg}
        }
    }

    # handle old tcl/tk versions too
    set delimiter {}
    if {8.0 <= ${tcl_version}} {
        set delimiter {::}
    }

    set fail 0
    foreach pkg [lsort ${pkgs}] {
        puts stdout ${pkg}
        catch {package require ${pkg} 1.0}
        # many of my namespaces use an initialisation proc for autoloading
        #   and dynamically creating procs at run time 
        catch {${delimiter}${pkg}${delimiter}Init}
        # get namespace commands and variables
        set cmds [lsort [info commands ${delimiter}${pkg}${delimiter}*]]
        set local_cmds {}
        regsub -all -- "^${delimiter}${pkg}${delimiter}" ${cmds} {} local_cmds
        # check namespace procs now
        foreach cmd $sgxRSM(tests) {
            incr fail [::sgxRSM::${cmd} ${cmds} [concat ${cmds} ${local_cmds} [info commands]]]
        }
        # remove package
        catch {package forget ${pkg}}
    }
    return ${fail}
}

# ------------------------------------------------------------------------------
# ::sgxRSM::_rsm_ --
#                     resource standard metrics
# Arguments:
#        procs:       list of procs to test
#        cmds:        list of cmds for call statistics
# Results:
#        integer      total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxRSM::_rsm_ {procs {cmds {}}} {
    # --------------------------------------------------------------------------
    global sgxRSM

    if {![llength ${cmds}]} {
        # use all interpreter commands
        set cmds [info commands]
    }

    set fail 0
    foreach name {Calls CCM FPM FRPM FC LOC eLOC _scum_} {
        set rsm_proc(${name}) 0
        set rsm_total(${name}) 0
    }

    puts stdout ""
    foreach cmd ${procs} {
        if {[catch {
            set proc_body [info body ${cmd}]
        }]} {
            # skip, it's not a proc
            continue
        }

        # command count metrics
        incr rsm_proc(CCM)

        # count blank lines, comment lines, appended comments
        set bl 0
        set cl 0
        set cl_appended 0
        set body ""
        foreach line [split ${proc_body} "\n\r"] {
            if {[regsub -all -- {^\W*#.*$} ${line} {} UNUSED]} {
                # a raw comment line
                incr cl
                continue
            }
            if {[regsub -all -- {;\W*#.*$} ${line} {} line]} {
                # a comment appended to a line (not best style)
                incr cl_appended
            } elseif {[regsub -all -- {^\W*$} ${line} {} UNUSED]} {
                # a blank line
                incr bl
            }
            append body "\n${line}"
        }

        # lines of code: loc
        set rsm_proc(LOC) [regsub -all -- "\[\n\r\]" ${proc_body} {} UNUSED]

        # effective lines of code: eloc
        set rsm_proc(eLOC) [expr {$rsm_proc(LOC) - ${cl} - ${bl}}]

        # source code understandability metrics
        #         ((All Comment Line Count + Blank Line Count) / Effective Line Count ) * 100
        set rsm_proc(SCUM) 0
        if {$rsm_proc(eLOC)} {
            set rsm_proc(_scum_) [expr {${cl} + ${cl_appended} + ${bl}}]
            set rsm_proc(SCUM) [format {%3.1f} [expr {$rsm_proc(_scum_) /(0.01 * $rsm_proc(eLOC))}]]
        }

        # language keyword metrics
        foreach call ${cmds} {
            set call_count [regsub -all -- "\\W${call}\\W" ${body} {} UNUSED]
            if {${call_count}} {
                set LKM(${call}) ${call_count}
                incr rsm_proc(Calls) ${call_count}
            }
        }

        # function interface complexity
        #         function parameter metric
        set rsm_proc(FPM) [llength [info args ${cmd}]]
        #         function return/exit point metric
        set rsm_proc(FRPM) [regsub -all -- {\W(return|exit)\W} ${body} {} UNUSED]
        #         function calls
        set rsm_proc(FC) [array size LKM]

        # create a single proc report
        puts stdout "# ------------------------------------------------"
        puts stdout "Proc :\t${cmd}"
        puts stdout "FPM  :\t$rsm_proc(FPM)\tFRPM :\t$rsm_proc(FRPM)\tFC   :\t$rsm_proc(FC)"
        puts stdout "LOC  :\t$rsm_proc(LOC)\teLOC :\t$rsm_proc(eLOC)\tSCUM :\t$rsm_proc(SCUM)%"

        foreach name {Calls CCM FPM FRPM FC LOC eLOC _scum_} {
            incr rsm_total(${name}) $rsm_proc(${name})
        }
        if {$sgxRSM(LKM,proc)} {
            puts stdout "\nCalls:\t$rsm_proc(Calls)"
        }
        foreach call [lsort [array names LKM]] {
            if {$sgxRSM(LKM,proc)} {
                puts stdout "\t$LKM(${call})\t${call}"
            }
            if {![info exists rsm_total_LKM(${call})]} {
                set rsm_total_LKM(${call}) 0
            }
            incr rsm_total_LKM(${call}) $LKM(${call})
        }
        foreach {rule descr} [join $sgxRSM(rules,proc)] {
            if {[catch {set active [eval expr ${rule}]} err]} {
                puts stderr "proc\trule invalid:\t${descr}\t${rule}\t${err}"
                incr fail
            } elseif {!${active}} {
                puts stderr "proc\t${cmd}\t${descr}\t${rule}\t[subst ${rule}]"
                incr fail
            }
        }
        # reset all counters
        set rsm_proc(Calls) 0
        set rsm_proc(CCM) 0
        catch {unset LKM}
    }

    # create a summary report
    if {${rsm_total(CCM)} && $sgxRSM(summary)} {
        puts stdout "# ------------------------------------------------\n"
        puts stdout "# ================================================"
        puts stdout "RSM Summary"

        puts stdout "\ntotal values\nCCM  :\t$rsm_total(CCM)"
        puts stdout "FPM  :\t$rsm_total(FPM)\tFRPM :\t$rsm_total(FRPM)\tFC   :\t[llength\
          [array names rsm_total_LKM]]"
        puts stdout "LOC  :\t$rsm_total(LOC)\teLOC :\t$rsm_total(eLOC)"

        puts stdout "\nmean values"
        foreach name {FPM FRPM FC LOC eLOC} {
            set rsm_mean(${name}) [format {%5.1f} [expr {1.0 * $rsm_total(${name}) /\
              $rsm_total(CCM)}]]
        }
        set rsm_mean(SCUM) [format {%5.1f} [expr {$rsm_total(_scum_) /(0.01 * $rsm_total(eLOC))}]]
        puts stdout "FPM  :\t$rsm_mean(FPM)\tFRPM :\t$rsm_mean(FRPM)\tFC   :\t$rsm_mean(FC)"
        puts stdout "LOC  :\t$rsm_mean(LOC)\teLOC :\t$rsm_mean(eLOC)\tSCUM :\t$rsm_mean(SCUM)%"

        if {$sgxRSM(LKM,total)} {
            puts stdout "\nCalls:\t$rsm_total(Calls)"
            foreach call [lsort [array names rsm_total_LKM]] {
                puts stdout "\t$rsm_total_LKM(${call})\t${call}"
            }
        }
        foreach mode {total mean} {
            foreach {rule descr} [join $sgxRSM(rules,${mode})] {
                if {[catch {set active [eval expr ${rule}]} err]} {
                    puts stderr "${mode}\trule invalid:\t${descr}\t${rule}\t${err}"
                    incr fail
                } elseif {!${active}} {
                    puts stderr "${mode}\t${descr}\t${rule}\t[subst ${rule}]"
                    incr fail
                }
            }
        }
        puts stdout "# ================================================\n"
    }
    catch {unset rsm_total_LKM}
    return ${fail}
}

# ==============================================================================

# examples of usage (self):

# first initialize rules
# ::sgxRSM::Init [file join .. test sgxRSM.tcl]

# now call an example
# ::sgxRSM::_Files_ [file join .. test test_tvc.dat]
# ::sgxRSM::_Pkgs_ ::sgxRSM
# ::sgxRSM::_rsm_ {::sgxRSM::_Files_ ::sgxRSM::_Pkgs_ ::sgxRSM::_rsm_}

# ==============================================================================

